#!/bin/sh

readonly THIS_PATH="$(readlink -e "$(dirname "$0")")"

. "${THIS_PATH}/systemd-unit-path.inc"

readonly INSTALL_LOG_FILEPATH="${1:-/dev/stdout}"

readonly PRODUCT_NAME='kesl'
readonly VAR_DIRECTORY='/var/opt/kaspersky/kesl/'
readonly SHARED_DIRECTORY='/opt/kaspersky/kesl/shared/'
readonly SUPERVISOR_FILENAME='kesl-supervisor'
readonly SUPERVISOR_FILEPATH='/etc/init.d/kesl-supervisor'
readonly SYSTEMD_SERVICE_FILENAME='kesl-supervisor.service'
readonly PACKAGE_CURRENT_DIRLINK='/var/opt/kaspersky/kesl/install-current'
readonly INI_FILEPATH='/var/opt/kaspersky/kesl/common/kesl.ini'
readonly SYSTEMD_SLEEP_DIRECTORY='/lib/systemd/system-sleep/'
readonly PM_UTILS_SLEEP_DIRECTORY='/etc/pm/sleep.d/'
readonly SYSTEMD_SLEEP_HOOK_FILENAME='kesl_hook.sh'
readonly PM_UTILS_SLEEP_HOOK_FILENAME='74_kesl_hook.sh'
readonly CONTROL_FILEPATH='/opt/kaspersky/kesl/bin/kesl-control'
readonly CONTROL_FILENAME='kesl-control'
readonly INSTALL_GUI_FILEPATH='/opt/kaspersky/kesl/libexec/install_gui.sh'

PKG_TYPE=''
SYSTEMD_UNIT_PATH=''


Die()
{
    local res=1
    if [ "$1" -eq "$1" ] 2>/dev/null; then
        res=$1
        shift
    fi

    {
        echo
        /bin/echo -e "$(basename "$0"): ERROR: $@"
        echo
    } >&2

    exit ${res}
}

SaveToLog()
{
    echo "$1" >> "${INSTALL_LOG_FILEPATH}"
}

CreateSymlink()
{
    local lnk="${1:?}"
    local dst="${2:-${PACKAGE_CURRENT_DIRLINK%%/}$1}"

    if [ -h "${lnk}" ]; then
        unlink "${lnk}" || Die "cannot unlink symlink '${lnk}'"
    elif [ -f "${lnk}" ]; then
        unlink "${lnk}" || Die "cannot unlink file '${lnk}'"
    elif [ -d "${lnk}" ]; then
        rmdir "${lnk}" || Die "cannot remove directory '${lnk}'"
    fi

    ln -s "${dst}" "${lnk}" || Die "cannot create symlink '${lnk}' -> '${dst}'"

    SaveToLog "${lnk}"
}

CreateSymlinks()
{
    for f in "${THIS_PATH}/"*.symlinks; do
        local lnk
        while IFS='' read -r lnk; do
            CreateSymlink "${lnk}"
        done < "${f}"
    done
}

InstallControl()
{
    CreateSymlink "/usr/bin/${CONTROL_FILENAME}" "${PACKAGE_CURRENT_DIRLINK}${CONTROL_FILEPATH}"
}

InstallService()
{
    ln -sfT "${PACKAGE_CURRENT_DIRLINK}${SUPERVISOR_FILEPATH}" "${SUPERVISOR_FILEPATH}"

    GetSystemdUnitPath SYSTEMD_UNIT_PATH
    if [ -z "${SYSTEMD_UNIT_PATH}" ]; then
        if [ "${PKG_TYPE}" = "deb" ]; then
            update-rc.d "${SUPERVISOR_FILENAME}" start 50 2 3 4 5 . stop 30 0 1 6 . >/dev/null
        else
            chkconfig --add "${SUPERVISOR_FILENAME}"
        fi
    else
        cp -Lf "${SHARED_DIRECTORY}${SYSTEMD_SERVICE_FILENAME}" "${SYSTEMD_UNIT_PATH}/${SYSTEMD_SERVICE_FILENAME}"
        systemctl daemon-reload
        systemctl enable "${SYSTEMD_SERVICE_FILENAME}"
    fi
}

InstallSleepHook()
{
    if [ -d "${SYSTEMD_SLEEP_DIRECTORY}" ]; then
        CreateSymlink "${SYSTEMD_SLEEP_DIRECTORY}${SYSTEMD_SLEEP_HOOK_FILENAME}" "${PACKAGE_CURRENT_DIRLINK}/${SHARED_DIRECTORY}${SYSTEMD_SLEEP_HOOK_FILENAME}"
    fi

    if [ -d "${PM_UTILS_SLEEP_DIRECTORY}" ]; then
        CreateSymlink "${PM_UTILS_SLEEP_DIRECTORY}${PM_UTILS_SLEEP_HOOK_FILENAME}" "${PACKAGE_CURRENT_DIRLINK}/${SHARED_DIRECTORY}${PM_UTILS_SLEEP_HOOK_FILENAME}"
    fi
}

InstallGui()
{
    "${INSTALL_GUI_FILEPATH}"
}

Install()
{
    CreateSymlinks
    InstallControl
    InstallService
    InstallSleepHook
    InstallGui
}

DetectPackageType()
{
    if [ ! -f "${INI_FILEPATH}" ]; then
        Die "could not find '${INI_FILEPATH}' file"
    fi
    local pkgType
    if ! pkgType=$(grep '^\s*PackageType\s*=' "${INI_FILEPATH}"); then
        Die "could not detect installed package type"
    fi
    pkgType=$(echo "${pkgType}" | sed -e 's:\(.*=\s*\)\(\S*\)\(\s*\):\2:')
    case "${pkgType}" in
        "deb"|"rpm")
            PKG_TYPE="${pkgType}"
            ;;
        *)
            Die "'${pkgType}': unknown package type"
            ;;
    esac
}

Main()
{
    DetectPackageType
    Install
}

Main "$@"
exit $?
