package KL::AppBase;

use v5.10;
use strict;

use File::Copy qw(copy);
use Getopt::Long;
use KL::AppInfo;
use KL::Utils;
use POSIX qw(strftime setlocale);

my $ENV_AUTO_INSTALL_FILE = 'KLI_AUTO_INSTALL_FILE';
my $ENV_AUTO_INSTALL_SAVE = 'KLI_AUTO_INSTALL_SAVE';
my $LICENSE_KEY           = '1VAW9-NGK79-TBVMC-XY3G9';

BEGIN
{
    my $lang = substr(setlocale("LC_ALL"), 0, 2);
    my @LFS_LANGS = split(/\s*,\s*/, "en ru de fr");
    if (grep {$_ eq $lang} @LFS_LANGS)
    {
        eval "require KL::Localization::$lang; KL::Localization::$lang->import();";
    }
    else
    {
        require KL::Localization::en;
        KL::Localization::en->import();
    }
}

sub new
{
    my $class = shift;
    my $appid = shift;

    my $self = {
        'SETTINGS' => {},
        'OPTIONS'  => {},
        'DEFAULT_COMMAND' => 'postinstall',
    };

    $self->{'COMMANDS'} = [
        {
            'id' => 'postinstall',
            'spec' => 'postinstall',
            'sub' => 'do_postinstall',
            'descr' => $MSG_CMD_HELP_POSTINSTALL,
        },
        {
            'id' => 'autoinstall',
            'spec' => 'autoinstall=s',
            'sub' => 'auto_install',
            'descr' => $MSG_CMD_HELP_AUTOINSTALL,
        },
        {
            'id' => 'trigger',
            'spec' => 'trigger=s',
            'sub' => 'do_trigger',
        },
        {
            'id' => 'upgrade',
            'spec' => 'upgrade',
        },
        {
            'id' => 'debug',
            'spec' => 'd|debug',
        },
        {
            'id' => 'kernel_module',
            'spec' => 'kernel-module=s',
            'default' => undef,
        },
    ];

    $self->{'AUTOINSTALL'} = [];
    $self->{'IS_AUTOINSTALL'} = 0;

    $self->{APPINFO} = new KL::AppInfo($appid);

    KL::Utils::fatal(2, "Cannot read info file for application ID '%1'", $appid)
        if !defined $self->{APPINFO}->{VALUES};

    bless $self, $class;

    $self->{'SETUP'} = undef;
    $self->{'TITLE'} = $self->get_name_version();

    return $self;
}

sub appinfo
{
    my $self = shift;

    return $self->{APPINFO}->value(@_);
}

sub get_path
{
    my $self = shift;
    my $name = shift;

    return $self->{APPINFO}->value('paths', $name);
}

sub get_name_version
{
    my $self = shift;

    my $title = $self->appinfo('product', 'title') .
        ' version ' . $self->appinfo('product', 'version');
}

sub get_applog_path
{
    my $self = shift;
    return $self->get_path('logs') . "/install.log";
}

sub applog
{
    my $self = shift;
    my $message = shift;

    my $logfile = $self->get_applog_path();
    my $stamp = strftime('[%d/%m/%y %H:%M:%S]', localtime);

    my @lines = split /\n/, $message;
    foreach (@lines) {
        $_ = "$stamp $_";
    }

    $message = join("\n", @lines) . "\n";

    if ( open F, ">>$logfile" ) {
                print F $message;
                close F;
    } else {
        print STDERR "LOG: $message";
    }
}

sub options
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    if (defined $value) {
        $self->{OPTIONS}->{$name} = $value;
    }

    return $self->{OPTIONS}->{$name};
}

sub settings
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    if (defined $value) {
        $self->{SETTINGS}->{''}->{$name} = $value;
        $self->save_settings();

        KL::Utils::debug("AppBase::settings('$name','$value')");
    } else {
        $value = $self->{SETTINGS}->{''}->{$name};
        KL::Utils::debug("AppBase::settings('$name') = '$value'");
    }

    return $value;
}

sub remove_settings
{
    my $self = shift;

    while (@_) {
        my $name = shift;

        delete $self->{SETTINGS}->{''}->{$name};
    }

    $self->save_settings();
}

sub load_settings
{
    my $self = shift;
    my $filename = $self->get_path('var');

    $self->{'SETTINGS'} = {};

    return if $filename eq '';

    $filename .= '/installer.dat';

    return 1 if ! -f $filename;

    my $data = KL::Utils::load_ini_file($filename);
    if (defined $data) {
        $self->{'SETTINGS'} = $data ;
        return 1;
    }

    return 0;
}

sub save_settings
{
    my $self = shift;
    my $filename = $self->get_path('installation_file');

    return if $filename eq '';

    return KL::Utils::save_ini_file($filename, $self->{'SETTINGS'});
}
sub on_agreement_acceptance {
    my $self = shift;
    my $agreement_type = shift;
    my $src = shift;
    my $dest;
    if ($agreement_type =~ /EULA/) {
        $dest = $self->get_path('agreed_eula');
    } elsif($agreement_type =~ /KSNUA/) {
        $dest = $self->get_path('agreed_ksnua');
    } else {
        return 0;
    }
    if (! -f $dest)
    {
        symlink $src, $dest;
    }
}

sub get_sutable_license_file
{
    my $self = shift;
    my $license_file_template = shift;
    my $system_lang;
    my ($ret, $locale_output) = KL::Utils::execute('locale');
    my $is_utf8 = 0;
    foreach my $line (split(/\n/, $locale_output)) {
    my ($locale_category, $value) = split(/=/, $line, 2);
    $value =~ s/^"(.+)"$/$1/;
        if ($locale_category =~ m/^LANG\s*$/)
    {
        ($system_lang, $value) = split(/_/, $value, 2);
        $is_utf8 = $value =~ m/.+(?i)UTf-?8$/;
        next;
    }
    if ($locale_category =~ m/^LC_MESSAGES\s*$/)
    {
        ($system_lang, $value) = split(/_/, $value, 2);
        $is_utf8 = 0;
        $is_utf8 = $value =~ m/.+(?i)UTf-?8$/;
        last;
    }
    }
    my $license_lang;
    foreach my $lang (split(/,/, $self->appinfo('product', 'license_langs')))
    {
        if ($system_lang eq $lang)
        {
            $license_lang = ${system_lang};
        }
    }
    my $default_license_language = (split(/,/, $self->appinfo('product', 'license_langs')))[0];
    if (defined($license_lang) and not $is_utf8 and not ($license_lang eq $default_license_language))
    {
    my $license_file = $license_file_template;
    $license_file =~ s/%LANG%/$license_lang/;
    KL::Utils::warning($MSG_CANNOT_SHOW_LICENSE_WITH_THE_CODEPAGE, $license_file);

    if (KL::Utils::ask_boolean('y', $MSG_SHOW_ENGLISH_AGREEMENT_QUESTION) == 0) {
        exit(1);
    }
    $license_lang = $default_license_language;
    }
    if (not defined($license_lang))
    {
    $license_lang = $default_license_language;
    }
    my $license_file = $license_file_template;
    $license_file =~ s/%LANG%/$license_lang/;
    return $license_file;
}

sub show_agreement {
    my $self = shift;
    my $agreement_type = shift;
    my ($HEADER_SHOW_AGREEMENT, $MSG_SHOW_AGREEMENT_BEGIN, $MSG_SHOW_AGREEMENT_FILE_PATH_MESSAGE, $MSG_SHOW_AGREEMENT_QUESTION,
        $MSG_SHOW_AGREEMENT_NOT_ACCEPTED, $agreement_file_template, $agreement_accepted, $agreement_setting);
    if ($agreement_type =~ /EULA/i) {
        $HEADER_SHOW_AGREEMENT = $HEADER_SHOW_EULA;
        $MSG_SHOW_AGREEMENT_BEGIN = $MSG_SHOW_EULA_BEGIN;
        $MSG_SHOW_AGREEMENT_FILE_PATH_MESSAGE = $MSG_SHOW_EULA_FILE_PATH_MESSAGE;
        $MSG_SHOW_AGREEMENT_QUESTION = $MSG_SHOW_EULA_QUESTION;
        $MSG_SHOW_AGREEMENT_NOT_ACCEPTED = $MSG_SHOW_EULA_NOT_ACCEPTED;
        $agreement_accepted = $self->get_path('agreed_eula');
        $agreement_file_template = $self->get_path('license_template');
        $agreement_setting = 'EULA_AGREED';
    } elsif($agreement_type =~ /KSNUA/i) {
        $HEADER_SHOW_AGREEMENT = $HEADER_SHOW_KSNUA;
        $MSG_SHOW_AGREEMENT_BEGIN = $MSG_SHOW_KSNUA_BEGIN;
        $MSG_SHOW_AGREEMENT_FILE_PATH_MESSAGE = $MSG_SHOW_KSNUA_FILE_PATH_MESSAGE;
        $MSG_SHOW_AGREEMENT_QUESTION = $MSG_SHOW_KSNUA_QUESTION;
        $MSG_SHOW_AGREEMENT_NOT_ACCEPTED = $MSG_SHOW_KSNUA_NOT_ACCEPTED;
        $agreement_accepted = $self->get_path('agreed_ksnua');
        $agreement_file_template = $self->get_path('ksn_license_template');
        $agreement_setting = 'KSNUA_AGREED';
    } else {
        return 1;
    }

    if (-f $agreement_accepted) {
        return 0;
    }

    my $agreement_file = $self->get_sutable_license_file($agreement_file_template);

    KL::Utils::header ($HEADER_SHOW_AGREEMENT);

    my $pager;
    my $message = $MSG_SHOW_AGREEMENT_BEGIN;

    if (defined($ENV{'PAGER'})) {
        $pager = $ENV{'PAGER'};
    } else {
        foreach my $bin ('/usr/bin/less', '/bin/more', '/bin/cat') {
            if ( -x  $bin ) {
                $pager = $bin;
                last;
            }
        }
    }

    if ($pager =~ /less/) {
        $pager = 'LESSCHARSET=utf-8 '.$pager;
        $pager .= ' --quit-at-eof';
        $message .= $MSG_SHOW_AGREEMENT_HOWTO_QUIT;
    } elsif ($pager =~ /more/) {
        $message .= $MSG_SHOW_AGREEMENT_HOWTO_QUIT;
    }

    $message .= $MSG_SHOW_AGREEMENT_END;

    KL::Utils::ask_question('', $message);

    my $rc = -1;
    if (defined $pager) {
        $rc = system("$pager $agreement_file");
    }

    if ($rc != 0) {
        open(EULA, "<$agreement_file") or
        KL::Utils::fatal(2, $MSG_SHOW_AGREEMENT_ERROR, $agreement_file, $!);

        while (<EULA>) {
            print;
        }
        close(EULA);
    }
    print "\n";

    KL::Utils::output($MSG_SHOW_AGREEMENT_FILE_PATH_MESSAGE, $agreement_file);

    print "\n";
    if (KL::Utils::ask_boolean('', $MSG_SHOW_AGREEMENT_QUESTION) == 0) {
        KL::Utils::output($MSG_SHOW_AGREEMENT_NOT_ACCEPTED, 0);
        return 2;
    }

    $self->settings($agreement_setting, 'Yes');
    $self->save_settings();
    $self->on_agreement_acceptance($agreement_type, $agreement_file);

    $self->ai_save($agreement_setting,  'Yes');
    return 0;
}


sub install_activation_impl
{
    my $self = shift;
    my $file = shift;

    KL::Utils::warning("App::install_activation(...) is not implemented!");

    return (0, "App::install_activation not implemented");
}

sub get_http_proxy
{
    my $self = shift;

    KL::Utils::warning("App::get_http_proxy(...) is not implemented!");

    return (128, "App::get_http_proxy not implemented");
}

sub set_http_proxy
{
    my $self = shift;
    my $value = shift;

    KL::Utils::warning("App::set_http_proxy(...) is not implemented!");

    return (1, "App::set_http_proxy not implemented");
}

sub run_updater
{
    my $self = shift;

    KL::Utils::warning("App::run_updater(...) is not implemented!");

    return 0;
}

sub set_autoupdates
{
    my $self = shift;
    my $enable = shift;

    KL::Utils::warning("App::enable_autoupdates(...) is not implemented!");

    return (0, "App::enable_autoupdates not implemented");
}

sub do_auto_install
{
    my $self = shift;

    KL::Utils::warning("App::do_auto_install(...) is not implemented!");

    return 0;
}

sub do_postinstall
{
    my $self = shift;

    KL::Utils::warning("App::do_postinstall(...) is not implemented!");

    return 0;
}

sub do_uninstall
{
    my $self = shift;
    my $upgrade = shift;

    KL::Utils::warning("App::do_uninstall(...) is not implemented!");

    return 0;
}



sub ai_mode
{
    my $self = shift;

    return $self->{'IS_AUTOINSTALL'};
}

sub ai_save
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    return if not defined($ENV{$ENV_AUTO_INSTALL_SAVE});

    my $filename = $ENV{$ENV_AUTO_INSTALL_SAVE};

    my $values = {};
    if ( -f $filename ) {
        $values = KL::Utils::load_ini_file($filename);
    }

    $values->{''}->{$name} = $value;

    if (!KL::Utils::save_ini_file($filename, $values)) {
        KL::Utils::warning($MSG_AI_SAVE_FAILED, $name, $value, $filename, $!);
    }
}

sub ai_load
{
    my $self = shift;
    my $filename = shift;

    $self->{AI_VALUES} = {};

    my $values = KL::Utils::load_ini_file($filename);

    if ($values) {
        $self->{AI_VALUES} = $values;
        return 1;
    }

    return 0;
}

sub ai_validate
{
    my $self = shift;

    my $errors = 0;

    foreach my $rule (@{$self->{'AUTOINSTALL'}})
    {
        my $id   = $rule->{'id'};
        my $type = $rule->{'type'};

        my $value = $self->ai_get_value($id);

        if (defined $value)
        {
            my($res, $error);
            if (defined $rule->{'list'})
            {
                ($res, $error) = $self->ai_validate_list($rule, $value);
                next if ($res);
            }

            if (defined $type)
            {
                my $function = "ai_validate_${type}";
                ($res, $error) = $self->$function($rule, $value);
                next if ($res);
            }

            $errors++;
            KL::Utils::error("'%2' is invalid value for '%1': %3.\n", $id,
                             $value, $error);
        }
        else
        {
            my $default = $rule->{'default'};
            if (defined($default))
            {
                $self->ai_set_value($id, $default);
            }
            else
            {
                KL::Utils::error("Required parameter '%1' is not set", $id);
                $errors++;
            }
        }
    }

    return ($errors == 0) ? 1 : 0;
}

sub ai_validate_mask
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $id = $rule->{'id'};
    my $mask = $rule->{'mask'};

    if ( $value =~ /$mask/) {
        return 1;
    }

    return (0, undef);
}

sub ai_validate_list
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $id = $rule->{'id'};
    my @list = @{$rule->{'list'}};

    foreach my $item (@list) {
        return 1 if lc($item) eq lc($value);
        $item = "'$item'";
    }

    my $valid_list = pop @list;
    if (scalar(@list)) {
        $valid_list = join(', ', @list) . ' or ' . $valid_list;
    }

    return (0, $valid_list);
}


sub ai_validate_dir
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    if (-d $value) {
        return 1;
    }

    return (0, "is not a path to directory");
}

sub ai_validate_activation
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;
    return 1;
}

sub ai_validate_file
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    if (-f $value) {
        return 1;
    }

    return (0, "is not a path to file");
}

sub ai_validate_url_or_abspath
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $protocol = '';

    if ($value =~ /^(\w+):\/\/(.*)$/)
    {
        $protocol = $1;
        $value = $2;

        if ( defined $rule->{'protocols'} ) {
            if ( not grep(/^${protocol}$/, @{$rule->{'protocols'}}) ) {
                return (0, "$protocol is not valid protocol");
            }
        }

        if ( ($protocol eq '') and defined($rule->{'protocol_requred'}) and $rule->{'protocol_requred'})
        {
            return (0, "Protocol must be specified for this URL");
        }

        return 1;
    }
    else
    {
        if ($value =~ /^\/.*/)
        {
            return 1;
        }
    }

    return (0, "Is not valid URL or absolute path");
}

sub ai_validate_url
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    my $protocol = '';

    if ($value =~ /^(\w+):\/\/(.*)$/ ) {
        $protocol = $1;
        $value = $2;

        if ( defined $rule->{'protocols'} ) {
            if ( not grep(/^${protocol}$/, @{$rule->{'protocols'}}) ) {
                return (0, "$protocol is not valid protocol");
            }
        }
    }

    if ( ($protocol eq '') and defined($rule->{'protocol_requred'}) and $rule->{'protocol_requred'}) {
        return (0, "protocol must be specified for this URL");
    }

    if ($value =~ /^((.*?)(:(.*?))?@)?(.*?)(:\d+)?/) {
        return 1;
    }

    return (0, "Is not valid URL");
}

sub ai_get_value
{
    my $self = shift;
    my $section = shift;
    my $name = shift;

    my $res;
    if (not defined $name)
    {
        $res = $self->{AI_VALUES}->{''}->{$section};
    }
    else
    {
        $res = $self->{AI_VALUES}->{$section}->{$name};
    }

    return $res;
}

sub ai_set_value
{
    my $self = shift;
    my $name = shift;
    my $value = shift;

    $self->{AI_VALUES}->{''}->{$name} = $value;
}

sub _ai_error
{
    my $self = shift;
    my $errorcode = shift;

    if ($errorcode < 70) {
        KL::Utils::fatal($errorcode, $MSG_AI_INSTALLATION_CANCELED);
    }

    if (not defined $self->{_AI_ERROR}) {
        $self->{_AI_ERROR} = $errorcode;
    }
}

sub auto_install
{
    my $self = shift;
    my $values = shift;

    $self->{IS_AUTOINSTALL} = 1;

    my $file = $values->{'autoinstall'};

    KL::Utils::warning($MSG_AI_EXECUTED_INFO, $file);

    if ( !$self->ai_load($file) ) {
        KL::Utils::fatal(2, $MSG_AI_LOAD_FAILED, $file);
    }

    if ($self->ai_validate()) {
        $self->{_AI_ERROR} = undef;

        my $res = $self->do_auto_install();

        if (defined $self->{_AI_ERROR}) {
            exit($self->{_AI_ERROR});
        }

        if (defined $self->{_AI_WARNINGS})
        {
            exit(71);
        }

        return $res;
    }

    KL::Utils::fatal(2, $MSG_AI_VALIDATION_FAILED);
}


sub trigger_install
{
    my $self = shift;
    my $values = shift;

    my $upgrade = defined $values->{'upgrade'};

    if (defined($ENV{$ENV_AUTO_INSTALL_FILE})) {
        my $file = $ENV{$ENV_AUTO_INSTALL_FILE};

        $self->show_app_title();

        $self->{IS_UPGRADE} = $upgrade;
        $values->{'autoinstall'} = $file;

        return $self->auto_install($values);
    }

    if ( defined $self->{'SETUP'} ) {
        my $title = $self->{APPINFO}->value('product', 'title');
        KL::Utils::output($MSG_NOT_CONFIGURED_NOTICE, $title, $self->{'SETUP'});
    }

    return 1;
}

sub trigger_uninstall
{
    my $self = shift;
    my $values = shift;

    my $upgrade = defined $values->{'upgrade'};

    return $self->do_uninstall($upgrade);
}

sub do_trigger
{
    my $self = shift;
    my $values = shift;

    my $trigger = $values->{'trigger'};

    if ($trigger eq 'install') {
        return $self->trigger_install($values);
    } elsif ($trigger eq 'uninstall') {
        return $self->trigger_uninstall($values);
    }

    KL::Utils::fatal(2, "Value for --trigger must be 'install' or 'uninstall', but not '%1'", $trigger);

    return 0;
}



sub do_install_activation
{
    my $self       = shift;
    my $activation = shift;

    my $success = $MSG_ACTIVATION_SUCCESS;
    my $fail    = $MSG_ACTIVATION_FAILED;

    if (!defined $activation || $activation eq '')
    {
        $activation = $LICENSE_KEY;
        $success    = $MSG_TRIAL_KEY_ACTIVATION_SUCCESS;
        $fail       = $MSG_TRIAL_KEY_ACTIVATION_FAILED;
    }

    use constant MAX_ATTEMPS    => 5;
    use constant SLEEP_TIMEOUT  => 2;
    use constant LOGIC_ERROR    => 65;
    use constant NETWORK_ERROR  => 67;

    my $attemps = 0;
    my $res;
    my $message;

    while (1)
    {
        ($res, $message) = $self->install_activation_impl($activation);
        if ($res == 0)
        {
            $self->ai_save('INSTALL_LICENSE', $activation);
            KL::Utils::output($success);
            return 1;
        }
        elsif ($res == NETWORK_ERROR)
        {
            $attemps++;
            if ($attemps >= MAX_ATTEMPS)
            {
                last;
            }

            sleep(SLEEP_TIMEOUT);
        }
        else
        {
            last;
        }
    }

    KL::Utils::output($fail, $activation, $message);
    if ($res == LOGIC_ERROR or $res == NETWORK_ERROR)
    {
        return $res;
    }
    return 0;
}

sub install_activation
{
    my $self = shift;
    my $is_license_present = shift;
    my $activation;

     while (1)
    {
        $activation = KL::Utils::ask_question('', $is_license_present
                                                  ? $MSG_ACTIVATION_ENTER_SKIP
                                                  : $MSG_ACTIVATION_ENTER_TRIAL);
        if ($activation eq '')
        {
            if ($is_license_present)
            {
                return 0;
            }
            else
            {
                last;
            }
        }

        $activation = KL::Utils::detildefy($activation);

        my($res, $message) = $self->install_activation_impl($activation);
        if ($res == 0)
        {
            $self->ai_save('INSTALL_LICENSE', $activation);
            KL::Utils::output($MSG_ACTIVATION_SUCCESS, $activation);
            return 1;
        }
        else
        {
            KL::Utils::output($MSG_ACTIVATION_FAILED, $activation, $message);
            print "\n";
        }

        return 0 if ($res < 64);
    }

    my $res = $self->do_install_activation($activation);
    if ($res != 1)
    {
        $res = 0;
    }
    return $res;
}

sub setup_update_source
{
    my $self = shift;
    my $task_id = shift;
    my $source = shift;

    if (!$task_id) {
        KL::Utils::error("Invalid function usage. task_id is mandatory argument\n\n");
        return 128;
    }

    KL::Utils::header ($HEADER_SETUP_UPDATE_SOURCE);

    if (defined $source) {
        my ($ret, $str) = $self->set_updater_source($source);

        if ($ret == 0)
        {
            KL::Utils::output($str);
            return $ret;
        }

        KL::Utils::error("%1\n\n", $str);
        return $ret;
    }

    my ($res, $current_options) = $self->get_task_settings_as_simple_hash($task_id);

    if ($res != 0 || ! defined $current_options)
    {
        KL::Utils::output($current_options);
        return $res;
    }

    my $default_value = "";
    if ($current_options->{'SourceType'} == 'Custom') {
        if ($current_options->{'CustomSources.item_0000'}) {
            if ('yes' eq lc($current_options->{'CustomSources.item_0000'}->{'Enabled'})) {
                $default_value = $current_options->{'CustomSources.item_0000'}->{'URL'};
            }
        }
    }
    if (! $default_value) {
        $default_value = 'KLServers';
    }

    while (1)
    {
        $source = KL::Utils::ask_question ($default_value, $MSG_SET_UPDATE_SOURCE);

        my ($ret, $str) = $self->set_updater_source($source);
        if ($ret == 0)
        {
            $self->ai_save('UPDATER_SOURCE', $source);
            return 0;
        }
        KL::Utils::warning("%1\n\n", $str);
        if ($ret != 2)
        {
            return $ret;
        }
    }
    return 128;
}

sub setup_proxy
{
    my $self = shift;
    my $value = shift;

    KL::Utils::header ($HEADER_SETUP_PROXY);

    if (defined $value) {
        my ($ok, $str) = $self->set_http_proxy($value);

        if ($ok == 0) {
            KL::Utils::output($MSG_SET_PROXY_UPDATED, $value);
            return 0;
        }

        KL::Utils::error("%1\n\n", $str);
        return $ok;
    }

    my ($res, $str) = $self->get_http_proxy();
    if ($res != 0)
    {
        KL::Utils::error("%1\n\n", $str);
        return $res;
    }

    my $default = $str // 'n';

    while ( 1 )
    {
        my $value = KL::Utils::ask_question($default, $MSG_SET_HTTP_PROXY);

        my $ai_value = $value;
        if ($value =~ /^n(o)?$/i)
        {
            $value = 'no';
            $ai_value = 'None';
        }

        ($res, $str) = $self->set_http_proxy($value);

        if ($res == 0)
        {
            $self->ai_save('PROXY_SERVER', $ai_value);
            return 0;
        }

        if ($res != 2)
        {
            KL::Utils::warning($MSG_UNEXPECTED_ERROR, $str);
            return $res;
        }

        KL::Utils::output($str);
    }

    return 128;
}

sub download_bases
{
    my $self = shift;
    my $auto = shift;

    KL::Utils::header($HEADER_DOWNLOAD_BASES);

    return $self->run_updater();
}

sub setup_autoupdates
{
    my $self = shift;
    my $auto = shift;

    my $enable = 'No';
    KL::Utils::header($HEADER_SETUP_AUTOUPDATES);
    if ($auto || KL::Utils::ask_boolean('y', $MSG_SET_AUTOUPDATES))
    {
        $enable = 'Yes';
    }
    my ($ok, $message) = $self->set_autoupdates($enable);

    if (not $ok)
    {
        KL::Utils::warning($MSG_SET_AUTOUPDATES_FAILED, $message);
        return 0;
    }

    if (lc($enable) eq 'yes')
    {
        KL::Utils::output($MSG_SET_AUTOUPDATES_SUCCESS, $message);
    }
    return 1;
}

sub show_app_title
{
    my $self = shift;

    KL::Utils::header($self->{'TITLE'});
}

sub do_show_help
{
    my $self = shift;

    print $MSG_USAGE_HELP;

    foreach my $option (@{$self->{'COMMANDS'}})
    {
        my $descr = $option->{'descr'};

        next if not defined $descr;

        if ($option->{'id'} eq 'autoinstall')
        {
            $descr = KL::Utils::translate_string($descr, '/opt/kaspersky/kesl/doc/autoinstall.ini');
        }
        else
        {
            $descr = KL::Utils::translate_string($descr);
        }

        $descr =~ s/\t/\n        /g;

        print "  $descr\n\n";
    }
}

sub parse_cmdline
{
    my $self = shift;

    my $values = {};

    my %specs;
    my @commands;
    my %functions;

    foreach my $option (@{$self->{'COMMANDS'}}) {
        my $id = $option->{'id'};
        my $spec = $option->{'spec'};
        my $default = $option->{'default'};

        $values->{$id} = $default;

        $specs{$spec} = \$values->{$id};

        if ($option->{'sub'})
        {
            push @commands, $id;
            $functions{$id} = $option->{'sub'};
        }
    }

    my $res = GetOptions (%specs);

    if (!$res) {
        $self->do_show_help();
        exit(1);
    }

    if ( !$values->{'trigger'} ) {
        $self->show_app_title();
    }

    KL::Utils::debug_mode(1) if $values->{'debug'};

    my ($command, $function);
    foreach my $cmd (@commands) {
        if ( defined $values->{$cmd} ) {
            $command = $cmd;
            last;
        }
    }

    $command = $self->{'DEFAULT_COMMAND'} if not defined $command;
    $function = $functions{$command};

    return $self->$function($values);
}

1;
