
/*
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2010
 * See License.txt for details
 *
 */


#ifndef KERNEL__KERNEL_H
#define KERNEL__KERNEL_H

#define KAVMON_IS_WRITTEN(flags)                                                    \
    (flags & (O_WRONLY | O_RDWR | O_APPEND | O_CREAT | O_TRUNC))

#if !defined(__linux__)
struct list_head
{
    struct list_head *next, *prev;
};
#endif

#define KLIST_HEAD_INIT(name)                                                       \
    {                                                                               \
        &(name), &(name)                                                            \
    }

#define KLIST_HEAD(name) struct list_head name = KLIST_HEAD_INIT(name)

#define KINIT_LIST_HEAD(ptr)                                                        \
    do                                                                              \
    {                                                                               \
        (ptr)->next = (ptr);                                                        \
        (ptr)->prev = (ptr);                                                        \
    } while (0)


#if !defined(__linux__)
/*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __list_add(struct list_head* new_, struct list_head* prev,
                                  struct list_head* next)
{
    next->prev = new_;
    new_->next = next;
    new_->prev = prev;
    prev->next = new_;
}

/**
 * list_add - add a new entry
 * @new: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
static __inline__ void list_add(struct list_head* new_, struct list_head* head)
{
    __list_add(new_, head, head->next);
}

/**
 * list_add_tail - add a new entry
 * @new: new entry to be added
 * @head: list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 */
static __inline__ void list_add_tail(struct list_head* new_, struct list_head* head)
{
    __list_add(new_, head->prev, head);
}

/*
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __list_del(struct list_head* prev, struct list_head* next)
{
    next->prev = prev;
    prev->next = next;
}

/**
 * list_del - deletes entry from list.
 * @entry: the element to delete from the list.
 * Note: list_empty on entry does not return true after this, the entry is in an
 * undefined state.
 */
static __inline__ void list_del(struct list_head* entry)
{
    __list_del(entry->prev, entry->next);
}

/**
 * list_del_init - deletes entry from list and reinitialize it.
 * @entry: the element to delete from the list.
 */
static __inline__ void list_del_init(struct list_head* entry)
{
    __list_del(entry->prev, entry->next);
    KINIT_LIST_HEAD(entry);
}

/**
 * list_empty - tests whether a list is empty
 * @head: the list to test.
 */
static __inline__ int list_empty(struct list_head* head)
{
    return head->next == head;
}

/**
 * list_splice - join two lists
 * @list: the new list to add.
 * @head: the place to add it in the first list.
 */
static __inline__ void list_splice(struct list_head* list, struct list_head* head)
{
    struct list_head* first = list->next;

    if (first != list)
    {
        struct list_head* last = list->prev;
        struct list_head* at   = head->next;

        first->prev = head;
        head->next  = first;

        last->next = at;
        at->prev   = last;
    }
}

/**
 * list_entry - get the struct for this entry
 * @ptr:	the &struct list_head pointer.
 * @type:	the type of the struct this is embedded in.
 * @member:	the name of the list_struct within the struct.
 */
#define list_entry(ptr, type, member)                                               \
    ((type*)((char*)(ptr) - (unsigned long)(&((type*)0)->member)))

/**
 * list_for_each	-	iterate over a list
 * @pos:	the &struct list_head to use as a loop counter.
 * @head:	the head for your list.
 */
#endif

#ifndef list_for_each
#define list_for_each(pos, head)                                                    \
    for (pos = (head)->next; pos != (head); pos = pos->next)
#endif

#ifdef LINUX_KERNEL24
#define timespec_t time_t
#define timespec_equal(a, b) ((*a) == (*b))
#else
#define timespec_t struct timespec
#endif

typedef struct
{
    dev_t dev;
    unsigned long ino;
    unsigned long p_ino;
    loff_t size;
    unsigned int n_links;
    timespec_t m_time;
    timespec_t c_time;
} file_info;

#define DEFAULT_MAX_FNAME_LEN 0x4000
#define DEFAULT_DENY_TOO_LONG_FNAME 1

typedef struct
{
    int cache_enable;
    int32_t version;
    uint32_t max_fname_len;
    int8_t deny_too_long_fname;
} monitor_config;

extern monitor_config m_config;

static __inline__ int FILE_EQUAL(const file_info* file1, const file_info* file2)
{
    return (file1->dev == file2->dev && file1->ino == file2->ino);
}

void Monitor_kernel_init(void);
void Monitor_kernel_finit(void);
int Monitor_kernel_register_file(const file_info* file);
void Monitor_kernel_unregister_file(const file_info* file);
int Monitor_kernel_trusted_file(const file_info* file);

int Monitor_kernel_register_proc(pid_t pid, const file_info* file);
void Monitor_kernel_unregister_proc(pid_t pid);
int Monitor_kernel_trusted_proc(pid_t pid, const file_info* file);

#ifdef __linux__
int Monitor_kernel_check_file(const file_info* info, check_req_data_t* reqd,
                              int* err);
int Monitor_kernel_check_file_sign(uid_t uid, gid_t gid, struct vfsmount* vfsmnt,
                                   struct dentry* dentry);
#else
typedef struct
{
    uid_t owner_uid;
    u_int32_t file_mode;
    const char* fstypename;
} file_info_ex;

int Monitor_kernel_check_file(uid_t uid, gid_t gid, pid_t pid, const char* filename,
                              const file_info* info, const file_info_ex* info_ex,
                              enum FileOpType filetype, int flags, int* err);
int Monitor_kernel_check_file_sign(uid_t uid, gid_t gid, const char* filename);
#endif

#endif // KERNEL__KERNEL_H
